#!/usr/bin/env python3

################################################################################
#
# Copyright (c) 2017, Perforce Software, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1.  Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#
# 2.  Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL PERFORCE SOFTWARE, INC. BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# = Date
#   
#   $Date: 2017/07/21 $
#
# = Description
#
#   Collect information about workspaces, task streams, labels and output them
#   in csv format:
#   Client,name,owner,last_update_date,last_access_date,nb_of_synced_files
#   Task,name,owner,last_update_date,last_access_date,nb_of_tmp_files
#   Label,name,owner,last_update_date,last_access_date,nb_of_tagged_files
#
#   To get the output sorted by last_access_date run: sort -t, -k5 output
#
#   db.have records related to a workspace with update/access date of 1970
#   can be safely deleted as the workspace spec has been deleted.
#
# = Usage
# 
#   oldClientsTasksLabels.py a_checkpoint|a_checkpoint.gz
#
################################################################################

from __future__ import print_function
import re
import time
import sys
import gzip
import mimetypes
import os
import math

def cmdUsage():
    sys.exit("Usage: oldClientsTasksLabels.py a_checkpoint|a_checkpoint.gz")

def main():
    if len(sys.argv) < 2:
       cmdUsage()
    clients = {}
    labels = {}
    streams = {}
    filename = sys.argv[1]
    ckpSize = os.path.getsize(filename)
    if mimetypes.guess_type(filename)[1] == 'gzip':
       ckp = gzip.open(filename, "rt")
       # rough estimate for uncompressed checkpoint:
       ckpSize *= 14 
    else:
       ckp = open(filename)
    readBytes = 0.0
    mod = 10.0
    clientSpecRE = re.compile('@pv@ [0-9]* @db.domain@ @(.*?)@ 99 @.*?@ @.*?@ @.*?@ @.*?@ @(.*?)@ ([0-9]*) ([0-9]*) [0-9]* @(.*)')
    haveRE = re.compile('@pv@ [0-9]* @db.have@ @//(.*?)/.*')
    streamRE = re.compile('@pv@ [0-9]* @db.stream@ @(//.*?/.*?)@ @.*@ @.*@ ([0-9]) ')
    streamSpecRE = re.compile('@pv@ [0-9]* @db.domain@ @(//.*?/.*?)@ 115 @@ @@ @@ @@ @(.*?)@ ([0-9]*) ([0-9]*) ')
    taskRE = re.compile('@pv@ [0-9]* @db.revtx@ @(//.*?/.*?)/.*?@')
    labelSpecRE = re.compile('@pv@ [0-9]* @db.domain@ @(.*?)@ 108 @@ @@ @@ @@ @(.*?)@ ([0-9]*) ([0-9]*) ')
    labelRE = re.compile('@pv@ [0-9]* @db.label@ @(.*?)@.*')
    print("0%...", end='', file=sys.stderr)
    sys.stderr.flush()
    for line in ckp:
        readBytes += len(line.rstrip())
        if math.floor((((readBytes / ckpSize) * 100.0) / mod)) == 1.0:
           print(str(int(mod)) + "%...", end='', file=sys.stderr)
           sys.stderr.flush()
           mod += 10
        match = clientSpecRE.search(line)
        if match:
           client = match.group(1)
           if client not in clients:
              clients[client] = {}
              clients[client]["have"] = 0
           clients[client]["owner"] = match.group(2)
           clients[client]["update"] = time.strftime("%Y/%m/%d %H:%M:%S", time.localtime(float(match.group(3))))   
           clients[client]["access"] = time.strftime("%Y/%m/%d %H:%M:%S", time.localtime(float(match.group(4))))   
        else:
          match = haveRE.search(line)
          if match:
             client = match.group(1)
             if client not in clients:
                clients[client] = {}
                clients[client]["have"] = 0
                clients[client]["owner"] = "unknown"
                clients[client]["update"] = time.strftime("%Y/%m/%d %H:%M:%S", time.localtime(0.0))
                clients[client]["access"] = time.strftime("%Y/%m/%d %H:%M:%S", time.localtime(0.0))
             clients[client]["have"] += 1
          else:
             match = streamRE.search(line)
             if match:
                stream = match.group(1)
                if stream not in streams:
                   streams[stream] = {}
                   streams[stream]["rev"] = 0
                streams[stream]["type"] = match.group(2)
             else:
                match = streamSpecRE.search(line)
                if match:
                   stream = match.group(1)
                   if stream not in streams:
                      streams[stream] = {}
                      streams[stream]["rev"] = 0
                   streams[stream]["owner"] = match.group(2)
                   streams[stream]["update"] = time.strftime("%Y/%m/%d %H:%M:%S", time.localtime(float(match.group(3))))   
                   streams[stream]["access"] = time.strftime("%Y/%m/%d %H:%M:%S", time.localtime(float(match.group(4))))   
                else:
                   match = taskRE.search(line)
                   if match: 
                      stream = match.group(1)
                      if stream not in streams:
                         streams[stream] = {}
                         streams[stream]["rev"] = 0
                         streams[stream]["owner"] = "unknown"
                         streams[stream]["update"] = time.strftime("%Y/%m/%d %H:%M:%S", time.localtime(0.0))
                         streams[stream]["access"] = time.strftime("%Y/%m/%d %H:%M:%S", time.localtime(0.0))
                      streams[stream]["rev"] += 1
                   else: 
                      match = labelSpecRE.search(line)
                      if match:
                         label = match.group(1)
                         if label not in labels:
                            labels[label] = {}
                            labels[label]["tag"] = 0
                         labels[label]["owner"] = match.group(2)
                         labels[label]["update"] = time.strftime("%Y/%m/%d %H:%M:%S", time.localtime(float(match.group(3))))   
                         labels[label]["access"] = time.strftime("%Y/%m/%d %H:%M:%S", time.localtime(float(match.group(4))))   
                      else:
                         match = labelRE.search(line)
                         if match:
                            label = match.group(1)
                            if label not in labels:
                               labels[label] = {}
                               labels[label]["tag"] = 0
                               labels[label]["owner"] = "unknown"
                               labels[label]["update"] = time.strftime("%Y/%m/%d %H:%M:%S", time.localtime(0.0))
                               labels[label]["access"] = time.strftime("%Y/%m/%d %H:%M:%S", time.localtime(0.0))
                            labels[label]["tag"] += 1
                         else:
                            match = re.search('.*@db.revcx@.*', line)
                            if match:
                               # stop, no need to look further
                               break
    ckp.close()
    print("100%", file=sys.stderr)
    sys.stderr.flush()
    for (key, value) in clients.items():
        print("Client," + key + "," + value["owner"] + "," + value["update"] +"," + value["access"] + "," + str(value["have"]))
    for (key, value) in streams.items():
        if value["type"] == "4":
           print("Task," + key + "," + value["owner"] + "," + value["update"] +"," + value["access"] + "," + str(value["rev"]))
    for (key, value) in labels.items():
        print("Label," + key + "," + value["owner"] + "," + value["update"] +"," + value["access"] + "," + str(value["tag"]))

if __name__ == '__main__':
   main()
